
      	(* Projet de programmation: outils de compression.
           fichier adapt.ml : codage adaptatif de Huffman  *)


   (* Description d'ensemble du programme:
      J'inscris tout d'abord la longueur (sur 4 octets) du fichier 
      compresser dans le futur fichier compress.

      Puis, je traite,  chaque caractere lu en entre, deux listes:
      !l_v et !l_p, de (int * int), le premier int symbolisant
      l'occurence partielle du caractre numrot par le 2me int.

      Dans !l_p, un couple particulier (0,256), sert  symboliser la
      sous-liste des caractres non encore utiliss.

      Dans !l_v, les caractres ont des occurences 1, afin de produire un
      arbre complet , mais la somme est limine lorsque l'on intgre ce
      sous-arbre  l'arbre principal.

      De plus, lorsque l_v est vide (ie quand tous les caractres ont t
      utiliss), on retire (0,256) de l_p, qui ne sert plus  rien.

      Donc,  chaque tape, la fonction "code" trouve la liste de 0/1
      corespondant au caractre trait, qui est envoye dans un tampon,
      qui est vid par paquets de 8 bits. *)

  

let ch_e   = ref std_in      (* pour le canal d'entre *)
and ch_s   = ref std_out     (* pour le canal de sortie *)
and l_v    = ref [(0,0)]     (* pour les caractres non encore utiliss *)
and l_p    = ref [(0,0)]     (* pour les autres et 256, symbolisant l'arbre
       	       	       	       	des caractres non encore utiliss *)
and tampon = ref [0];;   (* contiendra la suite de 0/1  regrouper en octets *)


    (* Produit la liste: [(1,255);(1,254);...;(1,0)]. *)

let rec fait_liste_vide =
  fun 0 -> [(1,0)]
    | i -> (1,i)::(fait_liste_vide (i-1));;


    (* Regarde si le caractre n est absent de la liste donne dcroissante. *)

let rec pas_element n =
  fun [] -> true
    | ((a,m)::l) -> if m=n
                    then false
                    else if m<n then true
               	    else pas_element n l;;


    (* Retire le caractere vide (0,256) de l_p, lorsque l_v est vide. *)

let test_tout_plein =
  fun [] ((0,256)::lp) -> ([],lp)
    | lv lp -> (lv,lp);;


    (* Retire le caractere n de la liste. *)
  
let rec retire n =
  fun ((a,m)::l) -> if m=n
                    then l
                    else (a,m)::(retire n l);;


    (* Insre un lment dans une liste de paires tries *)
    (* par ordre croissant du premier lment.           *)

let rec insere a =
  fun []     -> [a]
    | (b::l) -> if fst a <= fst b then a::b::l
                                  else b::(insere a l);;


    (* Retrie la liste (l_p) en incrmentant l'occurence de n.         *)
    (* Cette liste n'est jamais vide, donc pas de problme de matching *)

let rec modifie n =
  fun ((occ,m)::l) -> if m=n then insere (occ+1,n) l
      	       	       	     else (occ,m)::(modifie n l);;


    (* Reoit une liste trie de (occ,car), et un car, et rend le chemin    *)
    (* dans l'arbre de Huffman corespondant, sous forme d'une liste de 0/1. *)

let code m ll = rev (code_inv m ll)
  where rec code_inv n =
    fun ((occ_a,a)::(occ_b,b)::l) ->
              if      n=a then  0::(code_inv n (insere (occ_a+occ_b,n) l))
              else if n=b then  1::(code_inv n (insere (occ_a+occ_b,n) l))
      	       	          else  code_inv n (insere (occ_a+occ_b,a) l)
      | _ -> [];;


      (* Rend l'octet corespondant aux 8 1ers lments de !tampon. *)

let valeur () =
  let accu = ref 0
  in
    begin
      accu:=0;
      for j = 1 to 8 do
      	match !tampon
	with (tete::queue) ->  ( accu := 2 * !accu + tete; tampon := queue )
      done;
      !accu
    end;;


      (* Inscrit, dans le fichier comprss, l'octet corespondant aux 8      *)
      (* premiers lments du tampon, tant qu'il y a suffisamment d'lments.*)

let vide_tampon () =
  while (list_length !tampon >= 8)
  do
    (output_byte !ch_s (valeur()))
  done;;


    (* Rajoute  la fin de tampon la direction list du caractre *)
    (* qui vient d'tre trait, et le vide, si possible.         *) 

let tamponne l = (tampon := !tampon @ l); vide_tampon ();;


    (* Reoit le nouveau caractre. Et modifie l_p et l_v en consquence. *)

let suivant car =
  if (pas_element car !l_v)
  then ( tamponne (code car !l_p);
         l_p := modifie car !l_p )
  else ( tamponne ((code 256 !l_p) @ (code car !l_v));
         l_v := retire car !l_v;
         match test_tout_plein !l_v !l_p
         with  (l1,l2) -> (l_v:= l1; l_p:=l2);
         l_p := insere (1,car) !l_p );;


      (* S'occupe d'ouvrir les fichiers. *)

let ouvre nom =
  begin
    print_newline();
    print_string ("Compression en cours de: " ^ nom);
    print_newline ();
    ch_e := (open_in_bin nom);
    ch_s := (open_out_bin (nom ^ ".zzz"));
    seek_out !ch_s 0;
    seek_in !ch_e 0
  end;;


      (* S'occupe de fermer les fichiers. *)

let ferme nom =
  begin
    print_newline();
    print_string ("Compression de " ^ nom ^ " termine.");
    print_newline();
    close_in !ch_e;
    close_out !ch_s
  end;;


  (* Prend "nom_du_fichier__comprsser" en argument,    *)
  (* rend unit, en produisant le fichier comprss .zzz. *)

let comprime nom_fich =
  begin

      (* Initialisation. *)

    tampon := [];
    l_v := fait_liste_vide 255;
    l_p := [(0,256)];
    ouvre nom_fich;

      (* Stocke la longueur totale du fichier  comprsser *)
      (* en tte du fichier comprss.                     *)

    output_binary_int !ch_s (in_channel_length !ch_e);

      (* Passage sur le fichier  compresser, o chaque octet       *)
      (* est inser dans le tampon sous forme de sa direction list. *)

    for i=1 to (in_channel_length !ch_e) do
      suivant(input_byte !ch_e)
    done;

      (* Pour finir l'octet en cours, et fermer les fichiers. *)

    tampon := !tampon @ [0;0;0;0;0;0;0];
    vide_tampon ();
    ferme nom_fich

  end;;

(* Fin.  Pour lancer: comprime "nom_du_fichier";; *)
(* Il faut 1 minute (sur ss5) pour comprsser 7392 octets de texte. *)
